@@ -1,8 +1,12 @@
 #ifndef STRINGBUF_H
 #define STRINGBUF_H
-
-/* (c) 2017 Workware Systems Pty Ltd  -- All Rights Reserved */
-
+/**
+ * resizable string buffer
+ *
+ * (c) 2017-2020 Steve Bennett <steveb@workware.net.au>
+ *
+ * See utf8.c for licence details.
+ */
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -48,7 +52,7 @@ void sb_free(stringbuf *sb);
 stringbuf *sb_copy(stringbuf *sb);
 
 /**
- * Returns the length of the buffer.
+ * Returns the byte length of the buffer.
  * 
  * Returns 0 for both a NULL buffer and an empty buffer.
  */
@@ -98,7 +102,7 @@ static inline char *sb_str(const stringbuf *sb)
 }
 
 /**
- * Inserts the given string *before* (zero-based) 'index' in the stringbuf.
+ * Inserts the given string *before* (zero-based) byte 'index' in the stringbuf.
  * If index is past the end of the buffer, the string is appended,
  * just like sb_append()
  */
@@ -131,6 +135,13 @@ char *sb_to_string(stringbuf *sb);
 #endif
 
 #endif
+/**
+ * resizable string buffer
+ *
+ * (c) 2017-2020 Steve Bennett <steveb@workware.net.au>
+ *
+ * See utf8.c for licence details.
+ */
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
@@ -141,8 +152,10 @@ char *sb_to_string(stringbuf *sb);
 #include "stringbuf.h"
 #endif
 #ifdef USE_UTF8
+#ifndef UTF8_UTIL_H
 #include "utf8.h"
 #endif
+#endif
 
 #define SB_INCREMENT 200
 
@@ -180,7 +193,6 @@ void sb_append(stringbuf *sb, const char *str)
 
 void sb_append_len(stringbuf *sb, const char *str, int len)
 {
-	int utf8_strlen(const char *str, int bytelen);
 	if (sb->remaining < len + 1) {
 		sb_realloc(sb, sb->last + len + 1 + SB_INCREMENT);
 	}
@@ -443,11 +455,16 @@ void sb_clear(stringbuf *sb)
 #ifndef STRINGBUF_H
 #include "stringbuf.h"
 #endif
+#ifndef UTF8_UTIL_H
 #include "utf8.h"
+#endif
 
 #define LINENOISE_DEFAULT_HISTORY_MAX_LEN 100
 
+/* ctrl('A') -> 0x01 */
 #define ctrl(C) ((C) - '@')
+/* meta('a') ->  0xe1 */
+#define meta(C) ((C) | 0x80)
 
 /* Use -ve numbers here to co-exist with normal unicode chars */
 enum {
@@ -480,6 +497,8 @@ struct current {
     int cols;   /* Size of the window, in chars */
     int nrows;  /* How many rows are being used in multiline mode (>= 1) */
     int rpos;   /* The current row containing the cursor - multiline mode only */
+    int colsright; /* refreshLine() cached cols for insert_char() optimisation */
+    int colsleft;  /* refreshLine() cached cols for remove_char() optimisation */
     const char *prompt;
     stringbuf *capture; /* capture buffer, or NULL for none. Always null terminated */
     stringbuf *output;  /* used only during refreshLine() - output accumulator */
@@ -523,15 +542,17 @@ void linenoiseHistoryFree(void) {
     }
 }
 
+typedef enum {
+    EP_START,   /* looking for ESC */
+    EP_ESC,     /* looking for [ */
+    EP_DIGITS,  /* parsing digits */
+    EP_PROPS,   /* parsing digits or semicolons */
+    EP_END,     /* ok */
+    EP_ERROR,   /* error */
+} ep_state_t;
+
 struct esc_parser {
-    enum {
-        EP_START,   /* looking for ESC */
-        EP_ESC,     /* looking for [ */
-        EP_DIGITS,  /* parsing digits */
-        EP_PROPS,   /* parsing digits or semicolons */
-        EP_END,     /* ok */
-        EP_ERROR,   /* error */
-    } state;
+    ep_state_t state;
     int props[5];   /* properties are stored here */
     int maxprops;   /* size of the props[] array */
     int numprops;   /* number of properties found */
@@ -638,7 +659,7 @@ static void DRL_STR(const char *str)
     }
 }
 #else
-#define DRL(ARGS...)
+#define DRL(...)
 #define DRL_CHAR(ch)
 #define DRL_STR(str)
 #endif
@@ -807,7 +828,7 @@ static void cursorDown(struct current *current, int n)
 
 void linenoiseClearScreen(void)
 {
-    write(STDOUT_FILENO, "\x1b[H\x1b[2J", 7);
+    IGNORE_RC(write(STDOUT_FILENO, "\x1b[H\x1b[2J", 7));
 }
 
 /**
@@ -975,6 +996,10 @@ static int check_special(int fd)
     if (c < 0) {
         return CHAR_ESCAPE;
     }
+	else if (c >= 'a' && c <= 'z') {
+		/* esc-a => meta-a */
+		return meta(c);
+	}
 
     c2 = fd_read_char(fd, 50);
     if (c2 < 0) {
@@ -1246,43 +1271,52 @@ void linenoiseSetMultiLine(int enableml)
 }
 
 /* Helper of refreshSingleLine() and refreshMultiLine() to show hints
- * to the right of the prompt. */
-static void refreshShowHints(struct current *current, const char *buf, int availcols) {
+ * to the right of the prompt.
+ * Returns 1 if a hint was shown, or 0 if not
+ * If 'display' is 0, does no output. Just returns the appropriate return code.
+ */
+static int refreshShowHints(struct current *current, const char *buf, int availcols, int display)
+{
+    int rc = 0;
     if (showhints && hintsCallback && availcols > 0) {
         int bold = 0;
         int color = -1;
         char *hint = hintsCallback(buf, &color, &bold, hintsUserdata);
         if (hint) {
-            const char *pt;
-            if (bold == 1 && color == -1) color = 37;
-            if (bold || color > 0) {
-                int props[3] = { bold, color, 49 }; /* bold, color, fgnormal */
-                setOutputHighlight(current, props, 3);
-            }
-            DRL("<hint bold=%d,color=%d>", bold, color);
-            pt = hint;
-            while (*pt) {
-                int ch;
-                int n = utf8_tounicode(pt, &ch);
-                int width = char_display_width(ch);
-
-                if (width >= availcols) {
-                    DRL("<hinteol>");
-                    break;
+            rc = 1;
+            if (display) {
+                const char *pt;
+                if (bold == 1 && color == -1) color = 37;
+                if (bold || color > 0) {
+                    int props[3] = { bold, color, 49 }; /* bold, color, fgnormal */
+                    setOutputHighlight(current, props, 3);
                 }
-                DRL_CHAR(ch);
+                DRL("<hint bold=%d,color=%d>", bold, color);
+                pt = hint;
+                while (*pt) {
+                    int ch;
+                    int n = utf8_tounicode(pt, &ch);
+                    int width = char_display_width(ch);
 
-                availcols -= width;
-                outputChars(current, pt, n);
-                pt += n;
+                    if (width >= availcols) {
+                        DRL("<hinteol>");
+                        break;
+                    }
+                    DRL_CHAR(ch);
+
+                    availcols -= width;
+                    outputChars(current, pt, n);
+                    pt += n;
+                }
+                if (bold || color > 0) {
+                    clearOutputHighlight(current);
+                }
+                /* Call the function to free the hint returned. */
+                if (freeHintsCallback) freeHintsCallback(hint, hintsUserdata);
             }
-            if (bold || color > 0) {
-                clearOutputHighlight(current);
-            }
-            /* Call the function to free the hint returned. */
-            if (freeHintsCallback) freeHintsCallback(hint, hintsUserdata);
         }
     }
+    return rc;
 }
 
 #ifdef USE_TERMIOS
@@ -1303,6 +1337,7 @@ static void refreshEnd(struct current *current)
 
 static void refreshStartChars(struct current *current)
 {
+    (void)current;
 }
 
 static void refreshNewline(struct current *current)
@@ -1313,6 +1348,7 @@ static void refreshNewline(struct current *current)
 
 static void refreshEndChars(struct current *current)
 {
+    (void)current;
 }
 #endif
 
@@ -1327,6 +1363,7 @@ static void refreshLineAlt(struct current *current, const char *prompt, const ch
     int notecursor;
     int cursorcol = 0;
     int cursorrow = 0;
+    int hint;
     struct esc_parser parser;
 
 #ifdef DEBUG_REFRESHLINE
@@ -1504,10 +1541,22 @@ static void refreshLineAlt(struct current *current, const char *prompt, const ch
         cursorrow = displayrow;
     }
 
-    DRL("\nafter buf: displaycol=%d, displayrow=%d, cursorcol=%d, cursorrow=%d\n\n", displaycol, displayrow, cursorcol, cursorrow);
+    DRL("\nafter buf: displaycol=%d, displayrow=%d, cursorcol=%d, cursorrow=%d\n", displaycol, displayrow, cursorcol, cursorrow);
 
     /* (f) show hints */
-    refreshShowHints(current, buf, current->cols - displaycol);
+    hint = refreshShowHints(current, buf, current->cols - displaycol, 1);
+
+    /* Remember how many many cols are available for insert optimisation */
+    if (prompt == current->prompt && hint == 0) {
+        current->colsright = current->cols - displaycol;
+        current->colsleft = displaycol;
+    }
+    else {
+        /* Can't optimise */
+        current->colsright = 0;
+        current->colsleft = 0;
+    }
+    DRL("\nafter hints: colsleft=%d, colsright=%d\n\n", current->colsleft, current->colsright);
 
     refreshEndChars(current);
 
@@ -1552,16 +1601,48 @@ static int remove_char(struct current *current, int pos)
     if (pos >= 0 && pos < sb_chars(current->buf)) {
         int offset = utf8_index(sb_str(current->buf), pos);
         int nbytes = utf8_index(sb_str(current->buf) + offset, 1);
+        int rc = 1;
 
-        /* Note that we no longer try to optimise the remove-at-end case
-         * since control characters and wide characters mess
-         * up the simple count
+        /* Now we try to optimise in the simple but very common case that:
+         * - outputChars() can be used directly (not win32)
+         * - we are removing the char at EOL
+         * - the buffer is not empty
+         * - there are columns available to the left
+         * - the char being deleted is not a wide or utf-8 character
+         * - no hints are being shown
          */
+        if (current->output && current->pos == pos + 1 && current->pos == sb_chars(current->buf) && pos > 0) {
+#ifdef USE_UTF8
+            /* Could implement utf8_prev_len() but simplest just to not optimise this case */
+            char last = sb_str(current->buf)[offset];
+#else
+            char last = 0;
+#endif
+            if (current->colsleft > 0 && (last & 0x80) == 0) {
+                /* Have cols on the left and not a UTF-8 char or continuation */
+                /* Yes, can optimise */
+                current->colsleft--;
+                current->colsright++;
+                rc = 2;
+            }
+        }
+
         sb_delete(current->buf, offset, nbytes);
 
         if (current->pos > pos) {
             current->pos--;
         }
+        if (rc == 2) {
+            if (refreshShowHints(current, sb_str(current->buf), current->colsright, 0)) {
+                /* A hint needs to be shown, so can't optimise after all */
+                rc = 1;
+            }
+            else {
+                /* optimised output */
+                outputChars(current, "\b \b", 3);
+            }
+        }
+        return rc;
         return 1;
     }
     return 0;
@@ -1579,17 +1660,41 @@ static int insert_char(struct current *current, int pos, int ch)
         char buf[MAX_UTF8_LEN + 1];
         int offset = utf8_index(sb_str(current->buf), pos);
         int n = utf8_getchars(buf, ch);
+        int rc = 1;
 
         /* null terminate since sb_insert() requires it */
         buf[n] = 0;
 
-        /* Optimisation removed - see reason in remove_char() */
-
+        /* Now we try to optimise in the simple but very common case that:
+         * - outputChars() can be used directly (not win32)
+         * - we are inserting at EOL
+         * - there are enough columns available
+         * - no hints are being shown
+         */
+        if (current->output && pos == current->pos && pos == sb_chars(current->buf)) {
+            int width = char_display_width(ch);
+            if (current->colsright > width) {
+                /* Yes, can optimise */
+                current->colsright -= width;
+                current->colsleft -= width;
+                rc = 2;
+            }
+        }
         sb_insert(current->buf, offset, buf);
         if (current->pos >= pos) {
             current->pos++;
         }
-        return 1;
+        if (rc == 2) {
+            if (refreshShowHints(current, sb_str(current->buf), current->colsright, 0)) {
+                /* A hint needs to be shown, so can't optimise after all */
+                rc = 1;
+            }
+            else {
+                /* optimised output */
+                outputChars(current, buf, n);
+            }
+        }
+        return rc;
     }
     return 0;
 }
@@ -1605,7 +1710,7 @@ static void capture_chars(struct current *current, int pos, int nchars)
         int offset = utf8_index(sb_str(current->buf), pos);
         int nbytes = utf8_index(sb_str(current->buf) + offset, nchars);
 
-        if (nbytes) {
+        if (nbytes > 0) {
             if (current->capture) {
                 sb_clear(current->capture);
             }
@@ -1656,6 +1761,28 @@ static int insert_chars(struct current *current, int pos, const char *chars)
     return inserted;
 }
 
+static int skip_space_nonspace(struct current *current, int dir, int check_is_space)
+{
+    int moved = 0;
+    int checkoffset = (dir < 0) ? -1 : 0;
+    int limit = (dir < 0) ? 0 : sb_chars(current->buf);
+    while (current->pos != limit && (get_char(current, current->pos + checkoffset) == ' ') == check_is_space) {
+        current->pos += dir;
+        moved++;
+    }
+    return moved;
+}
+
+static int skip_space(struct current *current, int dir)
+{
+    return skip_space_nonspace(current, dir, 1);
+}
+
+static int skip_nonspace(struct current *current, int dir)
+{
+    return skip_space_nonspace(current, dir, 0);
+}
+
 /**
  * Returns the keycode to process, or 0 if none.
  */
@@ -1681,8 +1808,8 @@ static int reverseIncrementalSearch(struct current *current)
         c = fd_read(current);
         if (c == ctrl('H') || c == CHAR_DELETE) {
             if (rchars) {
-                int p = utf8_index(rbuf, --rchars);
-                rbuf[p] = 0;
+                int p_ind = utf8_index(rbuf, --rchars);
+                rbuf[p_ind] = 0;
                 rlen = strlen(rbuf);
             }
             continue;
@@ -1707,7 +1834,7 @@ static int reverseIncrementalSearch(struct current *current)
             searchdir = 1;
             skipsame = 1;
         }
-        else if (c >= ' ') {
+        else if (c >= ' ' && c <= '~') {
             /* >= here to allow for null terminator */
             if (rlen >= (int)sizeof(rbuf) - MAX_UTF8_LEN) {
                 continue;
@@ -1766,11 +1893,6 @@ static int reverseIncrementalSearch(struct current *current)
 static int linenoiseEdit(struct current *current) {
     int history_index = 0;
 
-    /* The latest history entry is always our current buffer, that
-     * initially is just an empty string. */
-    linenoiseHistoryAdd("");
-
-    set_current(current, "");
     refreshLine(current);
 
     while(1) {
@@ -1804,7 +1926,8 @@ static int linenoiseEdit(struct current *current) {
         switch(c) {
         case SPECIAL_NONE:
             break;
-        case '\r':    /* enter */
+        case '\r':    /* enter/CR */
+        case '\n':    /* LF */
             history_len--;
             free(history[history_len]);
             current->pos = sb_chars(current->buf);
@@ -1841,6 +1964,7 @@ static int linenoiseEdit(struct current *current) {
                 return -1;
             }
             /* Otherwise fall through to delete char to right of cursor */
+            /* fall-thru */
         case SPECIAL_DELETE:
             if (remove_char(current, current->pos) == 1) {
                 refreshLine(current);
@@ -1851,6 +1975,24 @@ static int linenoiseEdit(struct current *current) {
              * Future possibility: Toggle Insert/Overwrite Modes
              */
             break;
+        case meta('b'):    /* meta-b, move word left */
+            if (skip_nonspace(current, -1)) {
+                refreshLine(current);
+            }
+            else if (skip_space(current, -1)) {
+                skip_nonspace(current, -1);
+                refreshLine(current);
+            }
+            break;
+        case meta('f'):    /* meta-f, move word right */
+            if (skip_space(current, 1)) {
+                refreshLine(current);
+            }
+            else if (skip_nonspace(current, 1)) {
+                skip_space(current, 1);
+                refreshLine(current);
+            }
+            break;
         case ctrl('W'):    /* ctrl-w, delete word at left. save deleted chars */
             /* eat any spaces on the left */
             {
@@ -1973,6 +2115,10 @@ history_navigation:
             refreshLine(current);
             break;
         default:
+			if (c >= meta('a') && c <= meta('z')) {
+				/* Don't insert meta chars that are not bound */
+				break;
+			}
             /* Only tab is allowed without ^V */
             if (c == '\t' || c >= ' ') {
                 if (insert_char(current, current->pos, c) == 1) {
@@ -1988,6 +2134,7 @@ history_navigation:
 int linenoiseColumns(void)
 {
     struct current current;
+    current.output = NULL;
     enableRawMode (&current);
     getWindowSize (&current);
     disableRawMode (&current);
@@ -2029,7 +2176,7 @@ static stringbuf *sb_getline(FILE *fh)
     return sb;
 }
 
-char *linenoise(const char *prompt)
+char *linenoiseWithInitial(const char *prompt, const char *initial)
 {
     int count;
     struct current current;
@@ -2048,6 +2195,10 @@ char *linenoise(const char *prompt)
         current.nrows = 1;
         current.prompt = prompt;
 
+		/* The latest history entry is always our current buffer */
+		linenoiseHistoryAdd(initial);
+		set_current(&current, initial);
+
         count = linenoiseEdit(&current);
 
         disableRawMode(&current);
@@ -2063,6 +2214,11 @@ char *linenoise(const char *prompt)
     return sb ? sb_to_string(sb) : NULL;
 }
 
+char *linenoise(const char *prompt)
+{
+	return linenoiseWithInitial(prompt, "");
+}
+
 /* Using a circular buffer is smarter, but a bit more complex to handle. */
 int linenoiseHistoryAddAllocated(char *line) {
 
